#include <string>
#include <cmath>
#include <boost/dynamic_bitset.hpp>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
#include <set>
#include "my_getopt-1.5/getopt.h"
#include "Networks.h"
#include "Isomorphism.h"
#include "Modules.h"
#include "utils.h"

using namespace std;
using namespace boost;

/* In this code: "bead" represents "homologous domain or domain copy",
   "chrom_region" represents "domain".
*/

//#define DEBUG

// global variables
vector<string> nets_IDs;
vector<string> beads_IDs;
vector<string> chrom_regions_IDs;
map<int,vector<int> > chromregion2bead; // mapping between chrom_region idx and bead idx. Each chrom_region is mapped to 2 beads.
double min_density;
int min_freq;
bool outputIsomorphism=false; // default is false. This is to output all occurrences of a pattern in the active networks (where they occur).

/*-----------------------------------
command line parser's data structure
-----------------------------------*/
struct option longopts[] =
{
	/* name,                      has_arg,          flag, val */ /* longind */
	{ "help",                     no_argument,       0,   'h' }, /*       0 */
	{ "outputIsomorphisms",		  no_argument,       0,   'o' }, /*       1 */
	/* end-of-list marker */
	{ 0, 0, 0, 0 }
};

char *shortopts = "ho"; /* short options string */
int longind = 0; /* long option list index */

void print_usage() {
	cout << "USAGE: recoverPattern <option> [patterns file] [pattern index] [homologous domain IDs file] [domain IDs file] [graph IDs file] [graph files path] [graph file suffix] [min_density] [min_freq] [output file]" << endl;
	cout << "   option:" << endl;
	cout << "      -h    (--help) print help message" << endl;
	cout << "      -o    (--outputIsomorphism) output another file containing isomorphisms for all patterns' occurrences." << endl;
	cout << endl;
	cout << "Below is the description of the input files:" << endl;
	cout << "[patterns file]: input file, it is the PATTERN file generated by Step 3 \"Tensor-based frequent dense subgraph identification algorithm\", the software \"unweightedNetsTensor\". Its format refers to the file \"README_unweightedNetsTensor\"." << endl;
	cout << "[pattern index]: input number, it is the index (1-based) of the patterns in the [patterns file]. For example, if pattern index is 10, this software will recover the pattern of the 10-th line in the [patterns file]." << endl;
	cout << "[homologous domain IDs file]: input file, each homologous domain (we called \"bead\") has a string ID. This file is a list of all homologous domain IDs, each is in a line." << endl;
	cout << "[domain IDs file]: input file, each domain (we called \"chromosome region\") has a string ID. This file is a list of all domain IDs. Because each domain has two copies, each line (a chromosome region) also contains its copy indexes. It is a tab-delimited format is below:" << endl;
	cout << "\tdomain_ID\thomologous_domain/copy_A_index\thomologous_domain/copy_B_index" << endl;
	cout << "\t...\t...\t..." << endl;
	cout << "\t...\t...\t..." << endl;
	cout << "[original graph (CIG) IDs file]: input file, it is a list of graph (chromosome interaction graph, or CIG) string IDs. Each line is a graph ID." << endl;
	cout << "[original graph (CIG) path]: the directory of all input chromosome interaction graphs (CIG). In this directory, each graph file is named as \"graph_ID.net\". Each graph file format is below:" << endl;
	cout << "\tcolumn 1: Homologous domain 1 index (start from 1)." << endl;
	cout << "\tcolumn 2: Homologous domain 2 index." << endl;
	cout << "\tcolumn 3: Weight of the edge between Domains A and B. (It's 1 for unweighted networks)" << endl;
	cout << "[network file suffix]: the extension of the network filename, default is \".net\"." << endl;
	cout << "[min_density]: minimum subgraph density of a pattern" << endl;
	cout << "[min_freq]: minimum number of graphs that a pattern/module occurs" << endl;
	cout << "[output file]: the recovered pattern for the specified contracted pattern. Please use only the output file (using the option \"--outputIsomorphism\") with the extension name \"*.occurs\". Its format is tab-delimited and described below:" << endl;
	cout << "\tcolumn 1: number of domains in the recovered pattern" << endl;
	cout << "\tcolumn 2: number of graphs that the recovered pattern occurs" << endl;
	cout << "\tcolumn 3: the subgraph information in the 1st graph that the recovered pattern occurs. It is a triplet \"[network_index]:[subgraph_density]:[a list of bead indexes]\". Note that (1) bead index is the line number of this bead in the input \"bead IDs file\"; and (2) the list of bead indexes is in the format as [685,682,680,686,681,689,683,679,690]." << endl;
	cout << "\tcolumn 4: the subgraph information in the 2nd graph that the recovered pattern occurs. It is still a triplet as aforementioned." << endl;
	cout << "\tcolumn 5: ..." << endl;
	cout << "\tcolumn xxx: the subgraph information in the last graph that the recovered pattern occurs." << endl;
	cout << endl;
    cout << "version 1.1, February 01 2013" << endl;
    cout << "Please email XJZHOU@usc.edu or WEL@usc.edu" << endl;
    cout << "Further refers to http:\/\/zhoulab.usc.edu for help." << endl;
	cout << endl;
}

void parse_command_line(int argc, char * argv[])
{
	int opt; /* during argument parsing, opt contains the return value from getopt() */
	int longind = 0; /* long option list index */
	/* parse all options from the command line */
	while ((opt=getopt_long(argc, argv, shortopts, longopts, &longind)) != -1) {
		switch (opt) {
			case 'h': /* --help */
				print_usage(); /* 'parms_setting' is a global variable */
				exit(-1);
			case 'o': /* --outputIsomorphisms */
				outputIsomorphism = true;
				break;
			default: /* something unexpected has happened */
				print_usage();
				exit(-1);
		}
	}
}

int main(int argc, char* argv[]) {
	parse_command_line(argc, argv);

	if ((optind+10)>argc) {
		cerr << "Error: #input required parameters <> 10!" << endl;
		print_usage();
		exit(EXIT_FAILURE);
	}

	////////////////////
	//  read files
	////////////////////

	string module_file = argv[optind];
	optind++;
	int module_index = atoi(argv[optind]);
	optind++;
	Module mod_chrom_regions(module_file.c_str(), module_index);

	string bead_ids_file = argv[optind];
	optind++;
	cout << "read homologous domain IDs file: " << bead_ids_file << endl;
	beads_IDs = list_from_file_str(bead_ids_file);
	cout << "\t#homolougs_domains: " << beads_IDs.size() << endl;
#ifdef DEBUG
	print( beads_IDs, ", " );
#endif

	string chrom_region_ids_file = argv[optind];
	optind++;
	cout << "read domain IDs file: " << chrom_region_ids_file << endl;
	read_chromregions2beads_file(chrom_region_ids_file, chrom_regions_IDs, chromregion2bead);
	cout << "\t#domains: " << chrom_regions_IDs.size() << endl;

	string network_ids_file = argv[optind];
	optind++;
	cout << "read graphs IDs file: " << network_ids_file << endl;
	nets_IDs = list_from_file_str(network_ids_file);
	cout << "\t#graphs (CIGs): " << nets_IDs.size() << endl;
#ifdef DEBUG
	print( nets_IDs, "," );
#endif

	string nets_dir = argv[optind];
	optind++;
	string nets_file_suffix = argv[optind];
	optind++;

	min_density = atof(argv[optind]);
	optind++;
	cout << "min_density: " << min_density << endl;

	min_freq = atoi(argv[optind]);
	optind++;
	cout << "min_freq: " << min_freq << endl;

	string output_file = argv[optind];
	optind++;
	cout << "output_file: " << output_file << endl;

	////////////////////
	//  process data
	////////////////////
	vector<int> nets_module = mod_chrom_regions.getNetworks(); // indexes of networks
#ifdef DEBUG
	nets_module.resize(2); // debug
#endif
	vector<int> chrom_regions_module = mod_chrom_regions.getNodes(); // indexes of chrom_regions in the module
	vector<string> chrom_regions_IDs_module;
	get_strings(chrom_regions_IDs, chrom_regions_module, chrom_regions_IDs_module);// IDs of chrom_regions in the module
	// get chromosomes of chrom_regions in the module
	vector<string> chromsomes_module;
	get_chromosomes_of_chromregions(chrom_regions_IDs, chrom_regions_module, chromsomes_module);
	Isomorphism iso( chromsomes_module );
	// beads of chrom_regions in the module
	vector<int> beads; // these beads correspond to chrom_regions in the module. So #bead = 2*#chrom_regions_module
	for (int j=0; j<chrom_regions_module.size(); j++) {
		int chrom_region_idx = chrom_regions_module[j];
		for (int k=0; k<chromregion2bead[chrom_region_idx].size(); k++) {
			beads.push_back(chromregion2bead[chrom_region_idx][k]);
		}
	}
	int n_chrom_regions=chrom_regions_module.size();
	int n_beads=beads.size();
	if (n_beads != 2*n_chrom_regions) {
		cerr << "Error: #homologous_domains=" << n_beads << " != 2*#domains in the pattern. #domains=" << n_chrom_regions << endl;
		exit(EXIT_FAILURE);
	}
#ifdef DEBUG
	print(beads, ", "); // debug
#endif
	
	vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> > transactions(nets_module.size()); // each transaction is for a network
	vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> > X_transactions(nets_module.size()); // each transaction is for a network
	vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> > ISO_X_transactions(nets_module.size()); // each transaction is for a network
	if (nets_module.size()>nets_IDs.size()) {
		cerr << "Error: #graphs_in_pattern=" << nets_module.size() << " is greater than #all_graphs=" << nets_IDs.size() << "!" << endl;
		exit(EXIT_FAILURE);
	}
	int control=1;
	for (int i=0; i<nets_module.size(); i++) { // visit each network the module occurs.
		// if ((i+1)%control == 0) {
		//	cout << (i+1) << ", " << flush;
		// }
		int net_idx = nets_module[i];
		string net_file = nets_dir + nets_IDs[net_idx] + "." + nets_file_suffix;

		// fetch network that only involves with beads corresponding to the chrom_regions in the module.
		Network net(net_file, MULE_GRAPH_FORMAT, beads); // bead network (note: not chrom_region network)
#ifdef DEBUG
		string file = net_file+".debug";
		net.print(file);
#endif

		// check each isomorphism of this module
		set<ISOMORPHISM_CODING_NUMBER_TYPE> X, ISO_X;
		ISOMORPHISM_CODING_NUMBER_TYPE n_possible_beads_modules = (ISOMORPHISM_CODING_NUMBER_TYPE)pow(2.0,n_chrom_regions);
		for (ISOMORPHISM_CODING_NUMBER_TYPE x=0; x<n_possible_beads_modules; x++) { // check each "beads module" encoded by "x"
			dynamic_bitset<> x_bits(n_chrom_regions, x); // bead module is encoded by a binary code of "x" with bit length "n_chrom_regions".
			// get beads of this isomorphism
			vector<int> beads_iso;
			get_subnodes_of_nodes(x_bits, chrom_regions_module, chromregion2bead, beads_iso);
			double density = net.density_subgraph( beads_iso );
			if (density>=min_density) {
				vector<ISOMORPHISM_CODING_NUMBER_TYPE> iso_x;
				iso.getIsomorphisms(x, iso_x);
				X.insert(x); // insert x to X
				copy( iso_x.begin(), iso_x.end(), inserter( ISO_X, ISO_X.end() ) ); // insert iso_x to ISO_X
			}
		}
		// Please note above set "X" and "ISO_X" are already sorted, because numbers are inserted from small to large.
		set<ISOMORPHISM_CODING_NUMBER_TYPE> union_, intersect, result_ISO_X;
		set_union(X.begin(), X.end(), ISO_X.begin(), ISO_X.end(), inserter(union_, union_.end()));
		transactions[i] = union_;
		X_transactions[i] = X;
		set_intersection( X.begin(), X.end(), ISO_X.begin(), ISO_X.end(), inserter(intersect,intersect.begin()));
		if (!intersect.empty()) {
			set_difference(ISO_X.begin(), ISO_X.end(), intersect.begin(), intersect.end(), inserter(result_ISO_X, result_ISO_X.end()));
			ISO_X_transactions[i] = result_ISO_X;
		} else {
			ISO_X_transactions[i] = ISO_X;
		}
	}
	cout << endl;

	// get frequent item
	cout << "Get frequent subgraph" << endl;
	vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> > item2freq;
	get_freq_items(transactions, min_freq, item2freq);
	if (!item2freq.empty()) {
		// take the most frequent item "item2freq[0].first"
		ISOMORPHISM_CODING_NUMBER_TYPE freq_item = item2freq[0].first;
		int freq = item2freq[0].second;
		vector<string> beads_IDs_module;
		vector<int> beads_index_module;
		iso.getPatternOfIsomorphism(freq_item, chrom_regions_IDs_module, beads_IDs_module);
		iso.getPatternOfIsomorphism(freq_item, chromregion2bead, chrom_regions_module, beads_index_module);
		vector<int> occurrences, indexes_trans;
		get_occurrences(transactions, freq_item, indexes_trans);
		for (int i=0; i<indexes_trans.size(); i++) {
			occurrences.push_back(nets_module[indexes_trans[i]]); // occurrence is the network index
		}

		// write to modules file containing only beads
		ofstream out;
		out.open(output_file.c_str(), ios::app);
		out << beads_index_module.size() << "\t" << freq << "\t" << min_density << "\t";
		print( out, beads_index_module ); // output beads indexes
		out << "\t";
		print( out, occurrences ); // output beads indexes
		out << endl;
		out.close();

		// for display info to console
		cout << "The recovered pattern below most frequently occurs in " << freq << " graphs:" << endl;
		print( beads_IDs_module, ", " );
		print( beads_index_module, ", " );

		// write all isomorphisms of all networks this module occurs.
		if (outputIsomorphism) {
			string output_all_isomorphisms_file = output_file + ".occurs";
			out.open(output_all_isomorphisms_file.c_str(), ios::app);
			out << beads_index_module.size() << "\t" << freq << "\t" << min_density;
			for (int i=0; i<indexes_trans.size(); i++) { // indexes_trans[i] is index of transactions which the module occurs.
				int net_idx = occurrences[i];
				string net_file = nets_dir + nets_IDs[net_idx] + "." + nets_file_suffix;
				// fetch network that only involves with beads corresponding to the chrom_regions in the module.
				Network net(net_file, MULE_GRAPH_FORMAT, beads); // bead network (note: not chrom_region network)

				set<ISOMORPHISM_CODING_NUMBER_TYPE> X = X_transactions[indexes_trans[i]];
				vector<ISOMORPHISM_CODING_NUMBER_TYPE> iso_subgraphs;
				iso.getIsomorphisms(freq_item, iso_subgraphs);

				vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,double> > subgraph2density;

				// check if this freq_item's density, if it is a real subgraph
				if (X.find(freq_item) != X.end()) {
					dynamic_bitset<> x_bits(n_chrom_regions, freq_item); // bead module is encoded by a binary code of "x" with bit length "n_chrom_regions".
					vector<int> beads_module;
					get_subnodes_of_nodes(x_bits, chrom_regions_module, chromregion2bead, beads_module);
					double density = net.density_subgraph( beads_module );
					subgraph2density.push_back(make_pair(freq_item, density));
				}

				// check its density, if its isomorphism subgraph of this freq_item is a real subgraph
				for (int j=0; j<iso_subgraphs.size(); j++) {
					ISOMORPHISM_CODING_NUMBER_TYPE subgraph_iso = iso_subgraphs[j];
					if (X.find(subgraph_iso) != X.end()) {
						// get beads of this isomorphism
						dynamic_bitset<> x_bits(n_chrom_regions, subgraph_iso); // bead module is encoded by a binary code of "x" with bit length "n_chrom_regions".
						vector<int> beads_module;
						get_subnodes_of_nodes(x_bits, chrom_regions_module, chromregion2bead, beads_module);
						double density = net.density_subgraph( beads_module );
						subgraph2density.push_back(make_pair(subgraph_iso, density));
					}
				}

				sort(subgraph2density.begin(), subgraph2density.end(), descending_order_double);
				ISOMORPHISM_CODING_NUMBER_TYPE real_subgraph = subgraph2density[0].first;
				double density = subgraph2density[0].second;

				dynamic_bitset<> x_bits(n_chrom_regions, real_subgraph); // bead module is encoded by a binary code of "x" with bit length "n_chrom_regions".
				vector<int> beads_module;
				get_subnodes_of_nodes(x_bits, chrom_regions_module, chromregion2bead, beads_module);

				out << "\t" << net_idx+1 << ":" << density << ":";
				print( out, beads_module, "[", ",", "]" );
			}
			out << endl;
			out.close();
		}
	} else {
		ofstream out;
		out.open(output_file.c_str(), ios::app);
		out << endl; // leave a blank line for this module
		out.close();
		if (outputIsomorphism) {
			string output_all_isomorphisms_file = output_file + ".occurs";
			out.open(output_all_isomorphisms_file.c_str(), ios::app);
			out << endl; // leave a blank line for this module
			out.close();
		}
		// for display info to console
		cout << "None of recovered patterns frequently occurs in >= " << min_freq << " graphs!" << endl;
	}

	return 0;
}
